#!/usr/bin/env python

# Script to dissect UDP data exchange between ELV Max! software and the
# ELV Max! Cube. Can also extract firmware data from a capture of a
# firmware upgrade.
#
# This script requires Python 2.7 and the dpkt python library.
#
# Version: 2
#
# Copyright 2013 Matthijs Kooijman <matthijs@stdin.nl>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# Changelog
# 1: Initial version to dissect and extract pcap files
# 2: Add support for firmware.enc files and add more packet types

import os
import sys
import dpkt
import struct
import socket
import argparse
from binascii import hexlify, unhexlify

TYPE_REPLY = ">"
TYPE_IDENTIFY = "I"
TYPE_NETINFO = "N"
TYPE_BOOTLOADER = "B"
TYPE_STARTUPDATE = "U"
TYPE_WRITE = "W"
TYPE_NETCONF = "c"
TYPE_SET_NETCONF = "C"
TYPE_URL = "h"
TYPE_SET_URL = "H"
TYPE_REBOOT = "R"

types = {
    TYPE_REPLY: 'Reply',
    TYPE_IDENTIFY: 'Identify',
    TYPE_NETINFO: 'Get network info',
    TYPE_BOOTLOADER: 'Enter bootloader',
    TYPE_STARTUPDATE: 'Start firmware update',
    TYPE_WRITE: 'Write frame',
    TYPE_NETCONF: 'Get network config',
    TYPE_SET_NETCONF: 'Set network config',
    TYPE_URL: 'Get url',
    TYPE_SET_URL: 'Set url',
    TYPE_REBOOT: 'Reboot',
}

SENDER_SOFTWARE = 0x2a00
SENDER_FIRMWARE = 0x4170
SENDER_BOOTLOADER = 0x426c

senders = {
    SENDER_SOFTWARE:    "*0: Software",
    SENDER_FIRMWARE:    "Ap: Cube firmware",
    SENDER_BOOTLOADER:  "Bl: Cube bootloader",
}

directions = {
    SENDER_SOFTWARE:    "--->",
    SENDER_FIRMWARE:    "<---",
    SENDER_BOOTLOADER:  "<---",
}

write_frame_statuses = {
    0: "Fail",
    1: "OK",
    2: "Restart",
}

def process_firmware_enc(args):
    f = open(args.file, 'r')

    while True:
        framesize_hex = f.read(4)
        if not framesize_hex:
            return # EOF

        framesize = struct.unpack("!H", unhexlify(framesize_hex))[0]
        frame = unhexlify(f.read(framesize * 2))
        if args.action == 'dissect':
            sys.stdout.write("Frame of %d bytes: %s\n" % (framesize, hexlify(frame)))
        elif args.action == 'extract':
            sys.stdout.write(frame)

def process_pcap(args):
    f = open(args.file, 'r')

    pcap = dpkt.pcap.Reader(f)

    def pop(data, fmt):
        count = struct.calcsize(fmt)
        if (len(data) >= count):
            res = struct.unpack(fmt, str(data[0:count]))
            del data[0:count]
            return res
        else:
            raise Exception("Packet is not long enough. Need %d bytes, but have: %s" % (count, hexlify(data)))

    def dissect_netinfo(data):
        """ Dissect netinfo and netconf replies and set netconf request """
        (ip, gateway, netmask, dns1, dns2) = pop(data, "!4s4s4s4s4s")
        sys.stdout.write("\tIP address: %s\n" % socket.inet_ntoa(ip))
        sys.stdout.write("\tGateway address: %s\n" % socket.inet_ntoa(gateway))
        sys.stdout.write("\tNetmask: %s\n" % socket.inet_ntoa(netmask))
        sys.stdout.write("\tPrimary DNS: %s\n" % socket.inet_ntoa(dns1))
        sys.stdout.write("\tSecondary DNS: %s\n" % socket.inet_ntoa(dns2))
        # netinfo packet only has the above, netconf packet has two more bytes
        if data:
            (use_dhcp, use_enc) = pop(data, "!BB")
            sys.stdout.write("\tUse DHCP: %s (0x%x)\n" % ("Yes" if use_dhcp else "No", use_dhcp))
            sys.stdout.write("\tUse Encryption: %s (0x%x)\n" % ("Yes" if use_enc else "No", use_enc))

    def dissect_url(data):
        """ Dissect get url reply and set url request """
        (port, ) = pop(data, "!H")
        sys.stdout.write("\tPort: %d (0x%d)\n" % (port, port))
        (host, path) = data.split(",")
        del data[:]
        sys.stdout.write("\tHost: %s\n" % host)
        sys.stdout.write("\tPath: %s\n" % path)

    # Previous write data packet
    prev = -1
    # Seen last data packet ack?
    seen_last = False

    for ts, buf in pcap:
        eth = dpkt.ethernet.Ethernet(buf)
        if not isinstance(eth.data, dpkt.ip.IP) or not isinstance(eth.data.data, dpkt.udp.UDP):
            continue
        packet = eth.data.data.data
        data = bytearray(packet)

        (magic,) = pop(data, "!6s")

        if magic != "eQ3Max":
            continue

        (sender, serial, type) = pop(data, "!H10s1s")

        if args.action == 'dissect':
            try:
                sys.stdout.write("%s\n" % directions.get(sender, "-??-"))
                sys.stdout.write("\tBytes: %s%s\n" % (hexlify(packet[:40]), "..." if len(packet) > 40 else ""))
                sys.stdout.write("\tSender: 0x%x (%s)\n" % (sender, senders.get(sender, "??")))
                sys.stdout.write("\tSerial: %s\n" % serial)
                sys.stdout.write("\tType: %s (%s)\n" % (types.get(type, "Unknown"), type))

                if type == TYPE_REPLY: # Reply
                    (reply_to,) = pop(data, "!1s")
                    sys.stdout.write("\tReply to: %s (%s)\n" % (types.get(reply_to, "Unknown"), reply_to))
                    if reply_to == TYPE_IDENTIFY:
                        (unknown, rf, firmware) = pop(data, "!B3sH")
                        sys.stdout.write("\tUnknown value: 0x%s\n" % unknown)
                        sys.stdout.write("\tRF address: 0x%s\n" % hexlify(rf))
                        sys.stdout.write("\tFirmware version: %d (0x%x)\n" % (firmware, firmware))
                    elif reply_to == TYPE_NETINFO or reply_to == TYPE_NETCONF:
                        dissect_netinfo(data)
                    elif reply_to == TYPE_WRITE:
                        (status, num) = pop(data, "!BH")
                        sys.stdout.write("\tStatus: %s (0x%x)\n" % (write_frame_statuses[status], status))
                        sys.stdout.write("\tFrame number: %d\n" % num)
                    elif reply_to == TYPE_URL:
                        dissect_url(data)
                elif type == TYPE_WRITE: # Write data
                    (num, length) = pop(data, "!HH")
                    sys.stdout.write("\tFrame number: %d\n" % num)
                    if (num != prev + 1):
                        sys.stdout.write("\t%d frames%s missing!" % (num - prev - 1, "s" if num - prev - 1 > 1 else ""))
                    prev = num
                    sys.stdout.write("\tPayload length: %d bytes\n" % length)
                    sys.stdout.write("\tPayload: %s bytes\n" % hexlify(data))
                    del data[:]
                elif type == TYPE_SET_NETCONF:
                    dissect_netinfo(data)
                elif type == TYPE_SET_URL:
                    dissect_url(data)
                if data:
                    sys.stdout.write("\tUndissected data: %s\n" % hexlify(data))
            except Exception, e:
                sys.stdout.write(">>> Error dissecting packet: %s\n" % e)
        elif args.action == 'extract':
            if type == TYPE_WRITE:
                (num, length) = pop(data, "!HH")

                if seen_last:
                    sys.stderr.write(">>> More data frames after last frame? Aborting\n")
                    sys.exit(1)

                if num != prev + 1:
                    sys.stderr.write(">>> %d frame%s missing! Aborting\n" % (num - prev - 1, "s" if num - prev - 1 > 1 else ""))
                    sys.exit(1)
                prev = num

                sys.stdout.write(data)
            elif type == TYPE_REPLY:
                (reply_to,) = pop(data, "!B")
                if reply_to == TYPE_WRITE:
                    (status, num) = pop(data, "!BH")
                    # A status of 2 seems to indicate the final packet was
                    # received
                    if status == 2:
                        seen_last = True

    if args.action == 'extract' and not seen_last:
        sys.stderr.write(">>> Did not see the last data ack, firmware may be incomplete!\n")
        sys.exit(1)

def main():
    parser = argparse.ArgumentParser(description='Parse ELV Max! packet dump or firmware.enc file')
    parser.add_argument('-d', '--dissect', dest='action', action='store_const',
                       const='dissect', default=False, help='Show dissected packets')
    parser.add_argument('-e', '--extract', dest='action', action='store_const',
                       const='extract', default=False, help='Dump firmware data on stdout')
    parser.add_argument('-t', '--type', dest='type', choices=['pcap', 'enc'],
                       default=False, help='Type of file to process')
    parser.add_argument('file', metavar='file', help='pcap or firmware.enc file to examine')

    args = parser.parse_args()
    if not args.action:
        sys.stderr.write("Must specify either --dissect or --extract")
        sys.exit(1)

    if not args.type:
        filename, ext = os.path.splitext(args.file)
        if ext == '.pcap':
            args.type = 'pcap'
        elif ext == '.enc':
            args.type = 'enc'

    if not args.type:
        sys.stderr.write("No file type given and couldn't detect from filetype\n")
        sys.exit(1)

    if args.type == 'pcap':
        process_pcap(args)
    else:
        process_firmware_enc(args)

if __name__ == '__main__':
    main()
